#include <assert.h>
#include <windows.h>
#include "Inpainter.h"
#include <stdio.h>

HWND ProgressWindow;
unsigned ProgressMessage;
unsigned ProgressLParam;
void InpainterGetResult();
//#define cimg_display_type 0
#define cimg_use_visualcpp6
bool abortNow;
#define cimg_plugin "greycstoration.h"
#include "CImg.h"
using namespace cimg_library;


struct InpainterContext {
    InpainterContext() 
    {
    }
    CImg<unsigned char> mask;
    CImg<unsigned char> dest;
    float amplitude     ;
    unsigned int nb_iter      ;
    float sharpness     ;
    float anisotropy    ;
    float alpha         ;
    float sigma         ;
    bool fast_approx    ;
    float gauss_prec    ;
    float dl            ;
    float da            ;
    unsigned int interp ;
    unsigned int tile   ;
    unsigned int btile  ;
    float sdt           ;
    float sp            ;
    char *file_m        ;
    unsigned int dilate ;
    unsigned int init   ;
    unsigned iter;
    bool abort;
};

static InpainterContext* context = 0;

void
DebugShow( CImg<unsigned char>& img )
{
    CImgDisplay *disp = 0;
    const int sx = 2*CImgDisplay::screen_dimx()/3, sy = 2*CImgDisplay::screen_dimy()/3;
    int nwidth=img.dimx(), nheight=img.dimy();
    if (nwidth>sx) { nheight = nheight*sx/nwidth; nwidth = sx; }
    if (nheight>sy) { nwidth = nwidth*sy/nheight; nheight = sy; }
    disp = new CImgDisplay(img.get_resize(nwidth,nheight),"GREYCstoration",1);
    img.display(*disp);
}

void InpainterInit( unsigned width, unsigned height, unsigned char* srcImage, unsigned char* maskImage, 
        HWND progressWindow, DWORD progressMessage, DWORD progressLParam )
{
    assert(context == 0 );
    context = new InpainterContext;

    context->amplitude     = 20.0f;
    context->nb_iter      = 1000;
    context->sharpness     = 0.3f;
    context->anisotropy    = 1.0f;
    context->alpha         = 0.8f;
    context->sigma         = 2.0f;
    context->fast_approx    = true;
    context->gauss_prec    = 2.0f;
    context->dl            = 0.8f;
    context->da            = 30.0f;
    context->interp = 0;
    context->tile   = 512;
    context->btile  = 4;
    context->sdt           = 0.0f;
    context->sp            = 0.5f;
    context->file_m        = 0;
    context->dilate = 2;
    context->init   = 4;
    context->iter = 0;
    context->abort = false;
    ProgressWindow = progressWindow;
    ProgressMessage = progressMessage;
    ProgressLParam = progressLParam;

    context->mask.assign(width, height, 1, 3);
    for ( int y = 0; y < height; y++ ) {
        for( int x = 0; x < width; x++ ) {
            context->mask(x,y,0) = *maskImage;
            context->mask(x,y,1) = *maskImage;
            context->mask(x,y,2) = *maskImage++;
        }
    }

    context->dest.assign(width, height, 1, 3);
    for ( int y = 0; y < height; y++ ) {
        for( int x = 0; x < width; x++ ) {
            context->dest(x,y,0) = *srcImage++;
            context->dest(x,y,1) = *srcImage++;
            context->dest(x,y,2) = *srcImage++;
        }
    }

    //DebugShow(context->mask);
    //DebugShow(context->dest);

    context->mask.resize(context->dest.dimx(),context->dest.dimy(),1,1);
    if (context->dilate) context->mask.dilate(context->dilate);
    switch (context->init) {
        case 0 : { cimg_forXYV(context->dest,x,y,k) if (context->mask(x,y)) context->dest(x,y,k) = 0; } break;
        case 1 : { cimg_forXYV(context->dest,x,y,k) if (context->mask(x,y)) context->dest(x,y,k) = 255; } break;
        case 2 : { cimg_forXYV(context->dest,x,y,k) if (context->mask(x,y)) context->dest(x,y,k) = (unsigned char)(255*cimg::rand()); } break;
        case 3 : break;
        default: {
             typedef unsigned char uchar;
             CImg<unsigned char> tmask(context->mask), ntmask(tmask);
             CImg_3x3(M,uchar);
             CImg_3x3(I,unsigned char);
             while (CImgStats(ntmask,false).max>0) {
                 cimg_for3x3(tmask,x,y,0,0,M) if (Mcc && (!Mpc || !Mnc || !Mcp || !Mcn)) {
                     const float
                         ccp = Mcp?0.0f:1.0f, cpc = Mpc?0.0f:1.0f,
                             cnc = Mnc?0.0f:1.0f, ccn = Mcn?0.0f:1.0f,
                             csum = ccp + cpc + cnc + ccn;
                     cimg_forV(context->dest,k) {
                         cimg_get3x3(context->dest,x,y,0,k,I);
                         context->dest(x,y,k) = (unsigned char)((ccp*Icp + cpc*Ipc + cnc*Inc + ccn*Icn)/csum);
                     }
                     ntmask(x,y) = 0;
                 }
                 tmask = ntmask;
             }
         } 
         break;
    }

}

void InpainterGetProgress( DWORD* iter, DWORD* percent )
{
    const unsigned int progress = (unsigned int)context->dest.greycstoration_progress();
    *iter = context->iter;
    *percent = progress; 
}

unsigned char* resultPosition;

void InpainterGetResult()
{
    if (context->sdt>0) context->dest.sharpen(context->sdt,
            context->sp,context->alpha/3,context->sigma/3);
    context->dest.cut(cimg::type<unsigned char>::min(),cimg::type<unsigned char>::max());
    unsigned char* result = resultPosition;

    for ( int y = 0; y < context->dest.height; y++ ) {
        for( int x = 0; x < context->dest.width; x++ ) {
            *result++ = context->dest(x,y,0);
            *result++ = context->dest(x,y,1);
            *result++ = context->dest(x,y,2);
        }
    }

}

void InpainterNextIteration()
{
    assert(!context->dest.greycstoration_is_running());
    abortNow = false;
    context->dest.greycstoration_mask_run(context->mask,
            context->amplitude,
            context->sharpness,
            context->anisotropy,
            context->alpha,
            context->sigma,
            context->dl,
            context->da,
            context->gauss_prec,
            context->interp,
            context->fast_approx,
            context->tile,
            context->btile);
    printf("Started next iteration.\n");
}

void InpainterStop()
{
    if ( context ) {
        abortNow = true;
        while ( context->dest.greycstoration_is_running() ) {
            Sleep(200);
        }
        Sleep(200);
    }
    ProgressWindow = 0;
    delete context;
    context = 0;
}

int
main(int argc, char* argv[] )
{
    MSG msg;
    // command line should have a single argument -- 
    // the name of the shared memory object.
    if ( argc != 2 ) {
        fprintf(stderr, "Usage: inpainter.exe <name of shared memory object>\n");
        exit(-1);
    }

    const char* sharedMemName = argv[1];

    // open the shared memory object.
    // OpenFileMapping
    HANDLE hMem = OpenFileMapping( 
        FILE_MAP_READ | FILE_MAP_WRITE,
        FALSE,
        argv[1]
        );

    if ( hMem == 0 ) {
        fprintf(stderr, "OpenFileMapping failed: %d\n", GetLastError());
        exit(-1);
    }

    // MapViewOfFile
    InpainterSharedMemory_t* mem = (InpainterSharedMemory_t*) 
        MapViewOfFile(hMem, // handle to map object
               FILE_MAP_READ|FILE_MAP_WRITE,  // read/write permission
               0,                    
               0,                    
               0);                   

    if ( mem == 0 ) {
        fprintf(stderr, "MapViewOfFile failed: %d\n", GetLastError());
        exit(-1);
    }

    // Create a message queue.
    PeekMessage(&msg, NULL, WM_USER, WM_USER, PM_NOREMOVE);

    printf("Started.\n");

    InpainterInit( mem->width, mem->height, mem->data, mem->data + 3 * mem->width * mem->height,
        mem->hwnd, mem->iterationDoneMsg, mem->iterationParam );

    resultPosition = mem->data;

    InpainterNextIteration();

    printf("Waiting for message.\n");

    while(GetMessage(&msg, NULL, 0, 0)) { // never exit.
        printf("Got message: %d\n", msg.message);
        if ( msg.message == INPAINTER_NEXT_ITERATION ) {
            InpainterNextIteration();
        } else if ( msg.message == INPAINTER_GET_PROGRESS ) {
            DWORD iter, percent;
            InpainterGetProgress( &iter, &percent );
            PostMessage( mem->hwnd, mem->iterationProgress, mem->iterationParam, percent);
        }
    }

    printf("Exiting.\n");

    return 0;
}

